from .pluthon_ast import (
    ByteString,
    Integer,
    Bool,
    AST,
    Apply,
    Delay,
    Ite,
    Unit,
    Force,
    Pattern,
)
from .pluthon_sugar import (
    EqualsByteString,
    EqualsInteger,
    EqualsBool,
    PVar,
    PLambda,
    TraceError,
)
from dataclasses import field, dataclass
import typing

"""
Functional Data Structures that can store anything (as opposed to PlutusData derivatives)

Built on wrapped lambda terms
"""


def identity(x):
    return x


_BUILTIN_TYPE_MAP = {
    bytes: (ByteString, identity),
    str: (ByteString, lambda x: x.encode()),
    int: (Integer, identity),
    bool: (Bool, identity),
}

_EQUALS_MAP = {
    ByteString: EqualsByteString,
    Integer: EqualsInteger,
    Bool: EqualsBool,
}


def FunctionalMapExtend(
    old_statemonad: "FunctionalMap",
    names: typing.List[typing.Any],
    values: typing.List[AST],
) -> "FunctionalMap":
    additional_compares = Apply(
        old_statemonad,
        PVar("x"),
        PVar("def"),
    )
    for name, value in zip(names, values):
        keytype, transform = _BUILTIN_TYPE_MAP[type(name)]
        additional_compares = Ite(
            _EQUALS_MAP[keytype](PVar("x"), keytype(transform(name))),
            Delay(value),
            additional_compares,
        )
    return PLambda(
        ["x", "def"],
        additional_compares,
    )


class FunctionalMap(AST):
    # low level maps that only support equal type keys
    # but can store anything
    def __new__(
        cls, kv: typing.Optional[typing.Dict[typing.Any, AST]] = None
    ) -> "FunctionalMap":
        res = PLambda(["x", "def"], PVar("def"))
        if kv is not None:
            res = FunctionalMapExtend(res, kv.keys(), kv.values())
        return res


@dataclass
class FunctionalMapAccess(Pattern):
    m: AST
    k: AST
    default: AST = field(default_factory=lambda: TraceError("KeyError"))

    def compose(self):
        return Force(Apply(self.m, self.k, Delay(self.default)))


TOPRIMITIVEVALUE = b"0"


class WrappedValue(AST):
    def __new__(cls, uplc_obj: AST, attributes: FunctionalMap):
        updated_map = FunctionalMapExtend(attributes, [TOPRIMITIVEVALUE], [uplc_obj])
        return updated_map


def from_primitive(p: AST, attributes: AST):
    return WrappedValue(p, attributes)


def to_primitive(wv: WrappedValue):
    return Apply(wv, TOPRIMITIVEVALUE, Unit())


class FunctionalTuple(AST):
    def __new__(cls, *vs: AST) -> "FunctionalTuple":
        # idea: just construct a nested if/else comparison
        if not vs:
            return Unit()
        param_name = "__f__"
        return PLambda([param_name], Apply(PVar(param_name), *map(Delay, vs)))


class FunctionalTupleAccess(AST):
    def __new__(cls, tuple: AST, index: int, size: int):
        if size == 0:
            raise ValueError("Can not access elements of an empty tuple")
        return Apply(
            tuple, PLambda([f"v{i}" for i in range(size)], Force(PVar(f"v{index}")))
        )



from uplc.ast import Program as UPLCProgram

from .compiler_config import DEFAULT_CONFIG
from .optimize.constant_index_access_list import IndexAccessOptimizer
from .optimize.patterns import OncePatternReplacer, AllPatternReplacer
from .optimize.remove_trace import RemoveTrace
from .pluthon_ast import Program, AST
from .util import NoOp
from uplc.tools import compile as uplc_compile


def compile(
    x: Program,
    config=DEFAULT_CONFIG,
) -> UPLCProgram:
    """
    Returns compiled Pluto code in UPLC
    :param x: the program to compile
    """
    x_old_dumps = None
    x_new_dumps = x.dumps()
    # need to iterate so that pattern optimizations can be applied to patterns that are part of other patterns
    # we stop when a fixpoint is reached
    while x_new_dumps != x_old_dumps:
        x_old_dumps = x_new_dumps
        for step in [
            IndexAccessOptimizer() if config.constant_index_access_list else NoOp(),
            (
                (
                    OncePatternReplacer()
                    if config.iterative_unfold_patterns
                    else AllPatternReplacer()
                )
                if config.compress_patterns
                else NoOp()
            ),
            RemoveTrace() if config.remove_trace else NoOp(),
        ]:
            x = step.visit(x)
        x_new_dumps = x.dumps()
    x = x.compile()
    x = uplc_compile(
        x,
        config=config,
    )
    return x


def dumps(u: AST):
    return u.dumps()



from pluthon.pluthon_ast import Integer
from pluthon.pluthon_sugar import (
    IndexAccessList,
    ConstantIndexAccessList,
    NthField,
    ConstantNthField,
    IndexAccessListFast,
    ConstantIndexAccessListFast,
)
from pluthon.util import NodeTransformer


class IndexAccessOptimizer(NodeTransformer):
    """
    Replaces IndexAccesses to constants with ConstantIndexAccesses
    """

    def visit_IndexAccessList(self, node: IndexAccessList):
        if isinstance(node.i, Integer):
            return ConstantIndexAccessList(node.l, node.i.x)
        return node

    def visit_NthField(self, node: NthField):
        if isinstance(node.n, Integer):
            return ConstantNthField(node.d, node.n.x)
        return node

    def visit_IndexAccessListFast(self, node: IndexAccessListFast):
        if isinstance(node.i, Integer):
            return ConstantIndexAccessListFast(node.l, node.i.x)
        return node



from dataclasses import dataclass
from typing import Optional
from uplc import compiler_config as uplc_compiler_config


@dataclass(frozen=True)
class CompilationConfig(uplc_compiler_config.CompilationConfig):
    compress_patterns: Optional[bool] = None
    iterative_unfold_patterns: Optional[bool] = None
    constant_index_access_list: Optional[bool] = None
    remove_trace: Optional[bool] = None

    def update(
        self, other: Optional["CompilationConfig"] = None, **kwargs
    ) -> "CompilationConfig":
        own_dict = self.__dict__
        other_dict = other.__dict__ if isinstance(other, CompilationConfig) else kwargs
        return self.__class__(
            **{
                k: other_dict.get(k) if other_dict.get(k) is not None else own_dict[k]
                for k in set(own_dict.keys()) | set(other_dict.keys())
            }
        )


# The default configuration for the compiler
OPT_O0_CONFIG = (
    CompilationConfig()
    .update(uplc_compiler_config.OPT_O0_CONFIG)
    .update(
        compress_patterns=False,
        iterative_unfold_patterns=False,
    )
)
OPT_O1_CONFIG = (
    CompilationConfig()
    .update(uplc_compiler_config.OPT_O1_CONFIG)
    .update(OPT_O0_CONFIG)
    .update(
        compress_patterns=True,
        constant_index_access_list=True,
    )
)
OPT_O2_CONFIG = (
    CompilationConfig()
    .update(uplc_compiler_config.OPT_O2_CONFIG)
    .update(OPT_O1_CONFIG)
    .update()
)
OPT_O3_CONFIG = (
    CompilationConfig()
    .update(uplc_compiler_config.OPT_O3_CONFIG)
    .update(OPT_O2_CONFIG)
    .update(
        unique_variable_names=True,
        iterative_unfold_patterns=True,
        remove_trace=True,
    )
)
OPT_CONFIGS = [OPT_O0_CONFIG, OPT_O1_CONFIG, OPT_O2_CONFIG, OPT_O3_CONFIG]

DEFAULT_CONFIG = CompilationConfig().update(OPT_O1_CONFIG)

ARGPARSE_ARGS = {
    "compress_patterns": {
        "help": "Enables the compression of re-occurring code patterns. Can reduce memory and CPU steps but increases the size of the compiled contract.",
    },
    "iterative_unfold_patterns": {
        "help": "Enables iterative unfolding of patterns. Improves application of pattern optimization but is very slow.",
    },
    "constant_index_access_list": {
        "help": "Replace index accesses with constant parameters with optimized constant accesses. Can reduce memory and CPU steps but increases the size of the compiled contract.",
    },
    "remove_trace": {
        "help": "Removes trace calls from the compiled code. This will make debugging harder but reduces contract size.",
    },
}
for k in ARGPARSE_ARGS:
    assert (
        k in DEFAULT_CONFIG.__dict__
    ), f"Key {k} not found in CompilationConfig.__dict__"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

try:
    from .pluthon_ast import *  # noqa: F403
    from .pluthon_sugar import *  # noqa: F403
    from .pluthon_functional_data import *  # noqa: F403
    from .tools import *  # noqa: F403
    from .compiler_config import *  # noqa: F403
except ImportError as e:
    logging.error(
        "Error, trying to import dependencies. Should only occur upon package installation",
        exc_info=e,
    )

VERSION = (1, 3, 2)

__version__ = ".".join([str(i) for i in VERSION])
__author__ = "nielstron"
__author_email__ = "n.muendler@posteo.de"
__copyright__ = "Copyright (C) 2025 nielstron"
__license__ = "MIT"
__url__ = "https://github.com/opshin/pluthon"



from .pluthon_ast import (
    ByteString,
    Integer,
    Bool,
    AST,
    Apply,
    Delay,
    Ite,
    Unit,
    Force,
    Pattern,
    Lambda,
    Var,
    Let,
    Text,
    BuiltIn,
    Error,
)
from uplc import ast as uplc_ast
import typing
from dataclasses import field, dataclass

########## Pluto Abstractions that simplify handling complex structures ####################


def name_scheme_compatible_varname(x: str):
    return f"0{x}_"


def PVar(x: str):
    """
    A simple wrapper around Var to ensure adherence to the naming scheme
    :param x: name of the variable
    :return: variable for use with pluthon, adhering to the naming scheme
    """
    return Var(name_scheme_compatible_varname(x))


def PLambda(vars: typing.List[str], term: AST):
    """
    A simple wrapper around Lambda to ensure adherence to the naming scheme
    """
    return Lambda(list(map(name_scheme_compatible_varname, vars)), term)


def PLet(bindings: typing.List[typing.Tuple[str, AST]], term: AST):
    """
    A simple wrapper around Let to ensure adherence to the naming scheme
    """
    return Let([(name_scheme_compatible_varname(x), y) for x, y in bindings], term)


@dataclass
class RecFun(Pattern):
    x: AST

    def compose(self):
        return PLet(
            [("g", self.x)],
            Apply(PVar("g"), PVar("g")),
        )


@dataclass
class Not(Pattern):
    x: AST

    def compose(self):
        return IfThenElse(self.x, Bool(False), Bool(True))


@dataclass
class Iff(Pattern):
    x: AST
    y: AST

    def compose(self):
        return PLet([("y", self.y)], Ite(self.x, PVar("y"), Not(PVar("y"))))


@dataclass
class And(Pattern):
    x: AST
    y: AST

    def compose(self):
        return Ite(self.x, self.y, Bool(False))


@dataclass
class Or(Pattern):
    x: AST
    y: AST

    def compose(self):
        return Ite(self.x, Bool(True), self.y)


@dataclass
class Xor(Pattern):
    x: AST
    y: AST

    def compose(self):
        return PLet([("y", self.y)], Ite(self.x, Not(PVar("y")), PVar("y")))


@dataclass
class Implies(Pattern):
    x: AST
    y: AST

    def compose(self):
        return Ite(self.x, self.y, Bool(True))


def wrap_builtin_binop(b: uplc_ast.BuiltInFun):
    def wrapped(x: AST, y: AST):
        return Apply(BuiltIn(b), x, y)

    return wrapped


def wrap_builtin_binop_force(b: uplc_ast.BuiltInFun):
    def wrapped(x: AST, y: AST):
        return Apply(Force(BuiltIn(b)), x, y)

    return wrapped


def wrap_builtin_binop_force_force(b: uplc_ast.BuiltInFun):
    def wrapped(x: AST, y: AST):
        return Apply(Force(Force(BuiltIn(b))), x, y)

    return wrapped


def wrap_builtin_ternop(t: uplc_ast.BuiltInFun):
    def wrapped(x: AST, y: AST, z: AST):
        return Apply(BuiltIn(t), x, y, z)

    return wrapped


def wrap_builtin_ternop_force(t: uplc_ast.BuiltInFun):
    def wrapped(x: AST, y: AST, z: AST):
        return Apply(Force(BuiltIn(t)), x, y, z)

    return wrapped


def wrap_builtin_ternop_force_force(t: uplc_ast.BuiltInFun):
    def wrapped(x: AST, y: AST, z: AST):
        return Apply(Force(Force(BuiltIn(t))), x, y, z)

    return wrapped


def wrap_builtin_unop(u: uplc_ast.BuiltInFun):
    def wrapped(x: AST):
        return Apply(BuiltIn(u), x)

    return wrapped


def wrap_builtin_unop_force(u: uplc_ast.BuiltInFun):
    def wrapped(x: AST):
        return Apply(Force(BuiltIn(u)), x)

    return wrapped


def wrap_builtin_unop_force_force(u: uplc_ast.BuiltInFun):
    def wrapped(x: AST):
        return Apply(Force(Force(BuiltIn(u))), x)

    return wrapped


def wrap_builtin_hexop_force(u: uplc_ast.BuiltInFun):
    def wrapped(d: AST, v: AST, w: AST, x: AST, y: AST, z: AST):
        return Apply(Force(BuiltIn(u)), d, v, w, x, y, z)

    return wrapped


AddInteger = wrap_builtin_binop(uplc_ast.BuiltInFun.AddInteger)
SubtractInteger = wrap_builtin_binop(uplc_ast.BuiltInFun.SubtractInteger)
MultiplyInteger = wrap_builtin_binop(uplc_ast.BuiltInFun.MultiplyInteger)
DivideInteger = wrap_builtin_binop(uplc_ast.BuiltInFun.DivideInteger)
QuotientInteger = wrap_builtin_binop(uplc_ast.BuiltInFun.MultiplyInteger)
RemainderInteger = wrap_builtin_binop(uplc_ast.BuiltInFun.RemainderInteger)
ModInteger = wrap_builtin_binop(uplc_ast.BuiltInFun.ModInteger)
EqualsInteger = wrap_builtin_binop(uplc_ast.BuiltInFun.EqualsInteger)
LessThanInteger = wrap_builtin_binop(uplc_ast.BuiltInFun.LessThanInteger)
LessThanEqualsInteger = wrap_builtin_binop(uplc_ast.BuiltInFun.LessThanEqualsInteger)
AppendByteString = wrap_builtin_binop(uplc_ast.BuiltInFun.AppendByteString)
ConsByteString = wrap_builtin_binop(uplc_ast.BuiltInFun.ConsByteString)
SliceByteString = wrap_builtin_ternop(uplc_ast.BuiltInFun.SliceByteString)
LengthOfByteString = wrap_builtin_unop(uplc_ast.BuiltInFun.LengthOfByteString)
IndexByteString = wrap_builtin_binop(uplc_ast.BuiltInFun.IndexByteString)
EqualsByteString = wrap_builtin_binop(uplc_ast.BuiltInFun.EqualsByteString)
LessThanByteString = wrap_builtin_binop(uplc_ast.BuiltInFun.LessThanByteString)
LessThanEqualsByteString = wrap_builtin_binop(
    uplc_ast.BuiltInFun.LessThanEqualsByteString
)
Sha2_256 = wrap_builtin_unop(uplc_ast.BuiltInFun.Sha2_256)
Sha3_256 = wrap_builtin_unop(uplc_ast.BuiltInFun.Sha3_256)
Blake2b_256 = wrap_builtin_unop(uplc_ast.BuiltInFun.Blake2b_256)
VerifySignature = wrap_builtin_ternop(uplc_ast.BuiltInFun.VerifyEd25519Signature)
VerifyEd25519Signature = wrap_builtin_ternop(uplc_ast.BuiltInFun.VerifyEd25519Signature)
VerifyEcdsaSecp256k1Signature = wrap_builtin_ternop(
    uplc_ast.BuiltInFun.VerifyEcdsaSecp256k1Signature
)
VerifySchnorrSecp256k1Signature = wrap_builtin_ternop(
    uplc_ast.BuiltInFun.VerifySchnorrSecp256k1Signature
)
AppendString = wrap_builtin_binop(uplc_ast.BuiltInFun.AppendString)
EqualsString = wrap_builtin_binop(uplc_ast.BuiltInFun.EqualsString)
EncodeUtf8 = wrap_builtin_unop(uplc_ast.BuiltInFun.EncodeUtf8)
DecodeUtf8 = wrap_builtin_unop(uplc_ast.BuiltInFun.DecodeUtf8)
# Note: prefer using Ite
IfThenElse = wrap_builtin_ternop_force(uplc_ast.BuiltInFun.IfThenElse)
ChooseUnit = wrap_builtin_unop_force(uplc_ast.BuiltInFun.ChooseUnit)
Trace = wrap_builtin_binop_force(uplc_ast.BuiltInFun.Trace)
FstPair = wrap_builtin_unop_force_force(uplc_ast.BuiltInFun.FstPair)
SndPair = wrap_builtin_unop_force_force(uplc_ast.BuiltInFun.SndPair)
ChooseList = wrap_builtin_ternop_force_force(uplc_ast.BuiltInFun.ChooseList)
MkCons = wrap_builtin_binop_force(uplc_ast.BuiltInFun.MkCons)
HeadList = wrap_builtin_unop_force(uplc_ast.BuiltInFun.HeadList)
TailList = wrap_builtin_unop_force(uplc_ast.BuiltInFun.TailList)
NullList = wrap_builtin_unop_force(uplc_ast.BuiltInFun.NullList)
ChooseData = wrap_builtin_hexop_force(uplc_ast.BuiltInFun.ChooseData)
ConstrData = wrap_builtin_binop(uplc_ast.BuiltInFun.ConstrData)
MapData = wrap_builtin_unop(uplc_ast.BuiltInFun.MapData)
ListData = wrap_builtin_unop(uplc_ast.BuiltInFun.ListData)
IData = wrap_builtin_unop(uplc_ast.BuiltInFun.IData)
BData = wrap_builtin_unop(uplc_ast.BuiltInFun.BData)
UnConstrData = wrap_builtin_unop(uplc_ast.BuiltInFun.UnConstrData)
UnMapData = wrap_builtin_unop(uplc_ast.BuiltInFun.UnMapData)
UnListData = wrap_builtin_unop(uplc_ast.BuiltInFun.UnListData)
UnIData = wrap_builtin_unop(uplc_ast.BuiltInFun.UnIData)
UnBData = wrap_builtin_unop(uplc_ast.BuiltInFun.UnBData)
EqualsData = wrap_builtin_binop(uplc_ast.BuiltInFun.EqualsData)
MkPairData = wrap_builtin_binop(uplc_ast.BuiltInFun.MkPairData)
MkNilData = wrap_builtin_unop(uplc_ast.BuiltInFun.MkNilData)
MkNilPairData = wrap_builtin_unop(uplc_ast.BuiltInFun.MkNilPairData)
SerialiseData = wrap_builtin_unop(uplc_ast.BuiltInFun.SerialiseData)
Keccak_256 = wrap_builtin_unop(uplc_ast.BuiltInFun.Keccak_256)
Blake2b_224 = wrap_builtin_unop(uplc_ast.BuiltInFun.Blake2b_224)
IntegerToByteString = wrap_builtin_ternop(uplc_ast.BuiltInFun.IntegerToByteString)
ByteStringToInteger = wrap_builtin_binop(uplc_ast.BuiltInFun.ByteStringToInteger)
AndByteString = wrap_builtin_ternop(uplc_ast.BuiltInFun.AndByteString)
OrByteString = wrap_builtin_ternop(uplc_ast.BuiltInFun.OrByteString)
XorByteString = wrap_builtin_ternop(uplc_ast.BuiltInFun.XorByteString)
ComplementByteString = wrap_builtin_unop(uplc_ast.BuiltInFun.ComplementByteString)
ShiftByteString = wrap_builtin_binop(uplc_ast.BuiltInFun.ShiftByteString)
RotateByteString = wrap_builtin_binop(uplc_ast.BuiltInFun.RotateByteString)
CountSetBits = wrap_builtin_unop(uplc_ast.BuiltInFun.CountSetBits)
FindFirstSetBit = wrap_builtin_unop(uplc_ast.BuiltInFun.FindFirstSetBit)
ReadBit = wrap_builtin_binop(uplc_ast.BuiltInFun.ReadBit)
WriteBits = wrap_builtin_ternop(uplc_ast.BuiltInFun.WriteBits)
ReplicateByte = wrap_builtin_binop(uplc_ast.BuiltInFun.ReplicateByte)
Ripemd_160 = wrap_builtin_unop(uplc_ast.BuiltInFun.Ripemd_160)
Bls12_381_G1_Uncompress = wrap_builtin_unop(uplc_ast.BuiltInFun.Bls12_381_G1_Uncompress)
Bls12_381_G1_Compress = wrap_builtin_unop(uplc_ast.BuiltInFun.Bls12_381_G1_Compress)
Bls12_381_G1_Add = wrap_builtin_binop(uplc_ast.BuiltInFun.Bls12_381_G1_Add)
Bls12_381_G1_Neg = wrap_builtin_unop(uplc_ast.BuiltInFun.Bls12_381_G1_Neg)
Bls12_381_G1_ScalarMul = wrap_builtin_binop(uplc_ast.BuiltInFun.Bls12_381_G1_ScalarMul)
Bls12_381_G1_HashToGroup = wrap_builtin_binop(
    uplc_ast.BuiltInFun.Bls12_381_G1_HashToGroup
)
Bls12_381_G1_Equal = wrap_builtin_binop(uplc_ast.BuiltInFun.Bls12_381_G1_Equal)
Bls12_381_G2_Uncompress = wrap_builtin_unop(uplc_ast.BuiltInFun.Bls12_381_G2_Uncompress)
Bls12_381_G2_Compress = wrap_builtin_unop(uplc_ast.BuiltInFun.Bls12_381_G2_Compress)
Bls12_381_G2_Add = wrap_builtin_binop(uplc_ast.BuiltInFun.Bls12_381_G2_Add)
Bls12_381_G2_Neg = wrap_builtin_unop(uplc_ast.BuiltInFun.Bls12_381_G2_Neg)
Bls12_381_G2_ScalarMul = wrap_builtin_binop(uplc_ast.BuiltInFun.Bls12_381_G2_ScalarMul)
Bls12_381_G2_HashToGroup = wrap_builtin_binop(
    uplc_ast.BuiltInFun.Bls12_381_G2_HashToGroup
)
Bls12_381_G2_Equal = wrap_builtin_binop(uplc_ast.BuiltInFun.Bls12_381_G2_Equal)
Bls12_381_MillerLoop = wrap_builtin_binop(uplc_ast.BuiltInFun.Bls12_381_MillerLoop)
Bls12_381_MulMlResult = wrap_builtin_binop(uplc_ast.BuiltInFun.Bls12_381_MulMlResult)
Bls12_381_FinalVerify = wrap_builtin_ternop(uplc_ast.BuiltInFun.Bls12_381_FinalVerify)

NotByteString = ComplementByteString

# Generic Utils


def TraceConst(x: AST, y: AST):
    return Trace(Text(x), y)


def TraceError(x: str):
    return Apply(Error(), Trace(Text(x), Unit()))


@dataclass
class NotEqualsInteger(Pattern):
    a: AST
    b: AST

    def compose(self):
        return Not(EqualsInteger(self.a, self.b))


@dataclass
class Negate(Pattern):
    a: AST

    def compose(self):
        return SubtractInteger(Integer(0), self.a)


EqualsBool = Iff


# List Utils
@dataclass()
class EmptyList(AST):
    sample_value: uplc_ast.Constant
    _fields = []

    def compile(self) -> uplc_ast.AST:
        return uplc_ast.BuiltinList([], self.sample_value)

    def mk_nil_suffix(self):
        if isinstance(self.sample_value, uplc_ast.BuiltinPair):
            return f"Pair<{EmptyList(self.sample_value.l_value).mk_nil_suffix()}|{EmptyList(self.sample_value.r_value).mk_nil_suffix()}>"
        if isinstance(self.sample_value, uplc_ast.BuiltinList):
            return f"List{EmptyList(self.sample_value.sample_value).mk_nil_suffix()}"
        return self.sample_value.__class__.__name__

    def dumps(self) -> str:
        # Note: this is not a real builtin. Essentially, this is not pluto
        return f"MkNil{self.mk_nil_suffix()} ()"


def EmptyIntegerList():
    return EmptyList(uplc_ast.BuiltinInteger(0))


def EmptyByteStringList():
    return EmptyList(uplc_ast.BuiltinByteString(b""))


def EmptyTextList():
    return EmptyList(uplc_ast.BuiltinString(""))


def EmptyBoolList():
    return EmptyList(uplc_ast.BuiltinBool(False))


def EmptyUnitList():
    return EmptyList(uplc_ast.BuiltinUnit())


def EmptyListList(sample_value: uplc_ast.BuiltinList):
    return EmptyList(sample_value)


def EmptyPairList(sample_value: uplc_ast.BuiltinPair):
    return EmptyList(sample_value)


def EmptyDataList():
    # Create an empty list
    return MkNilData(Unit())


def EmptyDataPairList():
    # Create an empty list of pair type
    return MkNilPairData(Unit())


def IteNullList(lst: AST, i: AST, e: AST):
    """
    Ite based on whether a list is empty or not, choose over Ite(NullList(l), i, e) for performance reasons
    Executes i if lst is empty, e otherwise
    """
    # Careful: can not patternize due to use of delay/force
    return Force(
        ChooseList(
            lst,
            Delay(i),
            Delay(e),
        )
    )


# Prepend an element to a list
PrependList = MkCons


@dataclass
class SingleDataList(Pattern):
    x: AST

    def compose(self):
        return PrependList(self.x, EmptyDataList())


@dataclass
class SingleDataPairList(Pattern):
    x: AST

    def compose(self):
        return PrependList(self.x, EmptyDataPairList())


@dataclass
class FoldList(Pattern):
    """Left fold over a list l operator f: accumulator -> list_elem -> accumulator with initial value a"""

    lst: AST
    f: AST
    a: AST

    def compose(self):
        return Apply(
            PLambda(
                ["op"],
                RecFun(
                    PLambda(
                        ["fold", "xs", "a"],
                        IteNullList(
                            PVar("xs"),
                            PVar("a"),
                            Apply(
                                PVar("fold"),
                                PVar("fold"),
                                TailList(PVar("xs")),
                                Apply(PVar("op"), PVar("a"), HeadList(PVar("xs"))),
                            ),
                        ),
                    ),
                ),
            ),
            self.f,
            self.lst,
            self.a,
        )


@dataclass
class FoldListAbort(Pattern):
    """Left fold over a list l operator f: accumulator -> list_elem -> accumulator with initial value a. Aborts on predicate p(a) returning true"""

    lst: AST
    f: AST
    a: AST
    p: AST

    def compose(self):
        return Apply(
            PLambda(
                ["op", "pred"],
                RecFun(
                    PLambda(
                        ["fold", "xs", "a"],
                        Ite(
                            Apply(PVar("pred"), PVar("a")),
                            PVar("a"),
                            IteNullList(
                                PVar("xs"),
                                PVar("a"),
                                Apply(
                                    PVar("fold"),
                                    PVar("fold"),
                                    TailList(PVar("xs")),
                                    Apply(PVar("op"), PVar("a"), HeadList(PVar("xs"))),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
            self.f,
            self.p,
            self.lst,
            self.a,
        )


@dataclass
class RFoldList(Pattern):
    """Right fold over a list l operator f: accumulator -> list_elem -> accumulator with initial value a"""

    lst: AST
    f: AST
    a: AST

    def compose(self):
        return Apply(
            PLambda(
                ["op"],
                RecFun(
                    PLambda(
                        ["fold", "xs", "a"],
                        IteNullList(
                            PVar("xs"),
                            PVar("a"),
                            Apply(
                                PVar("op"),
                                Apply(
                                    PVar("fold"),
                                    PVar("fold"),
                                    TailList(PVar("xs")),
                                    PVar("a"),
                                ),
                                HeadList(PVar("xs")),
                            ),
                        ),
                    ),
                ),
            ),
            self.f,
            self.lst,
            self.a,
        )


_CONSTANT_INDEX_ACCESS_PATTERNS = {}
_CONSTANT_INDEX_ACCESS_PATTERNS_FAST = {}


def _NthConstantIndexAccessList(i: int):
    if i < 0:
        raise ValueError("Index must be non-negative")
    if _CONSTANT_INDEX_ACCESS_PATTERNS.get(i) is None:

        def assign_vars(self, lst: AST):
            self.lst = lst

        if i == 0:

            def compose(self):
                return Apply(
                    PLambda(
                        ["xs"],
                        IteNullList(
                            PVar("xs"), TraceError("IndexError"), HeadList(PVar("xs"))
                        ),
                    ),
                    self.lst,
                )

        else:

            def compose(self):
                return _NthConstantIndexAccessList(i - 1)(
                    Apply(
                        PLambda(
                            ["xs"],
                            IteNullList(
                                PVar("xs"),
                                TraceError("IndexError"),
                                TailList(PVar("xs")),
                            ),
                        ),
                        self.lst,
                    )
                )

        ConstantIndexAccessListPattern = type(
            f"ConstantIndexAccessListPattern_{i}",
            (Pattern,),
            {
                "__annotations__": {"lst": AST},
                "__init__": assign_vars,
                "compose": compose,
            },
        )
        ConstantIndexAccessListPattern = dataclass(ConstantIndexAccessListPattern)
        _CONSTANT_INDEX_ACCESS_PATTERNS[i] = ConstantIndexAccessListPattern
    return _CONSTANT_INDEX_ACCESS_PATTERNS[i]


def ConstantIndexAccessList(lst: AST, i: int):
    return _NthConstantIndexAccessList(i)(lst)


def _NthConstantIndexAccessListFast(i: int):
    if i < 0:
        raise ValueError("Index must be non-negative")
    if _CONSTANT_INDEX_ACCESS_PATTERNS_FAST.get(i) is None:

        def assign_vars(self, lst: AST):
            self.lst = lst

        if i == 0:

            def compose(self):
                return HeadList(self.lst)

        else:

            def compose(self):
                return _NthConstantIndexAccessListFast(i - 1)(TailList(self.lst))

        ConstantIndexAccessListPatternFast = type(
            f"ConstantIndexAccessListPatternFast_{i}",
            (Pattern,),
            {
                "__annotations__": {"lst": AST},
                "__init__": assign_vars,
                "compose": compose,
            },
        )
        ConstantIndexAccessListPatternFast = dataclass(
            ConstantIndexAccessListPatternFast
        )
        _CONSTANT_INDEX_ACCESS_PATTERNS_FAST[i] = ConstantIndexAccessListPatternFast
    return _CONSTANT_INDEX_ACCESS_PATTERNS_FAST[i]


def ConstantIndexAccessListFast(lst: AST, i: int):
    return _NthConstantIndexAccessListFast(i)(lst)


@dataclass
class IndexAccessList(Pattern):
    lst: AST
    i: AST

    def compose(self):
        return Apply(
            RecFun(
                PLambda(
                    ["f", "i", "xs"],
                    IteNullList(
                        PVar("xs"),
                        TraceError("IndexError"),
                        Ite(
                            EqualsInteger(PVar("i"), Integer(0)),
                            HeadList(PVar("xs")),
                            Apply(
                                PVar("f"),
                                PVar("f"),
                                SubtractInteger(PVar("i"), Integer(1)),
                                TailList(PVar("xs")),
                            ),
                        ),
                    ),
                ),
            ),
            self.i,
            self.lst,
        )


def n_times_taillist(a: AST, n: int):
    res = a
    for _ in range(n):
        res = TailList(res)
    return res


def IndexAccessListFast(step_size: int = 5):
    """
    Construct a pattern for step-size skip access
    """

    def compose(self):
        return Apply(
            PLet(
                [
                    (
                        "step_access",
                        RecFun(
                            PLambda(
                                ["f", "i", "xs"],
                                Ite(
                                    EqualsInteger(PVar("i"), Integer(0)),
                                    HeadList(PVar("xs")),
                                    Apply(
                                        PVar("f"),
                                        PVar("f"),
                                        SubtractInteger(PVar("i"), Integer(1)),
                                        TailList(PVar("xs")),
                                    ),
                                ),
                            ),
                        ),
                    ),
                    (
                        "skip_access",
                        RecFun(
                            PLambda(
                                ["f", "i", "xs"],
                                Ite(
                                    LessThanInteger(PVar("i"), Integer(step_size)),
                                    Apply(PVar("step_access"), PVar("i"), PVar("xs")),
                                    Apply(
                                        PVar("f"),
                                        PVar("f"),
                                        SubtractInteger(PVar("i"), Integer(step_size)),
                                        n_times_taillist(PVar("xs"), step_size),
                                    ),
                                ),
                            )
                        ),
                    ),
                ],
                Apply(PVar("skip_access"), self.i, self.lst),
            )
        )

    def assign_vars(self, lst: AST, i: AST):
        self.lst = lst
        self.i = i

    IndexAccessListFastType = type(
        f"IndexAccessListFastType_{step_size}",
        (Pattern,),
        {
            "__annotations__": {"lst": AST, "i": AST},
            "__init__": assign_vars,
            "compose": compose,
        },
    )
    IndexAccessListFastType = dataclass(IndexAccessListFastType)

    return IndexAccessListFastType


@dataclass
class Range(Pattern):
    limit: AST
    start: AST = field(default_factory=lambda: Integer(0))
    step: AST = field(default_factory=lambda: Integer(1))

    def compose(self):
        return Apply(
            PLambda(
                ["limit", "step"],
                RecFun(
                    PLambda(
                        ["f", "cur"],
                        Ite(
                            LessThanInteger(PVar("cur"), PVar("limit")),
                            PrependList(
                                PVar("cur"),
                                Apply(
                                    PVar("f"),
                                    PVar("f"),
                                    AddInteger(PVar("cur"), PVar("step")),
                                ),
                            ),
                            EmptyIntegerList(),
                        ),
                    )
                ),
            ),
            self.limit,
            self.step,
            self.start,
        )


@dataclass
class MapList(Pattern):
    """Apply a map function on each element in a list"""

    lst: AST
    m: AST = field(default_factory=lambda: PLambda(["x"], PVar("x")))
    empty_list: AST = field(default_factory=EmptyDataList)

    def compose(self):
        return Apply(
            PLambda(
                ["op"],
                RecFun(
                    PLambda(
                        ["map", "xs"],
                        IteNullList(
                            PVar("xs"),
                            self.empty_list,
                            PrependList(
                                Apply(PVar("op"), HeadList(PVar("xs"))),
                                Apply(
                                    PVar("map"),
                                    PVar("map"),
                                    TailList(PVar("xs")),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
            self.m,
            self.lst,
        )


@dataclass
class FindList(Pattern):
    """Returns the first element in the list where key evaluates to true - otherwise returns default"""

    lst: AST
    key: AST
    default: AST

    def compose(self):
        return Apply(
            PLambda(
                ["op"],
                RecFun(
                    PLambda(
                        ["f", "xs"],
                        IteNullList(
                            PVar("xs"),
                            self.default,
                            Ite(
                                Apply(PVar("op"), HeadList(PVar("xs"))),
                                HeadList(PVar("xs")),
                                Apply(
                                    PVar("f"),
                                    PVar("f"),
                                    TailList(PVar("xs")),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
            self.key,
            self.lst,
        )


@dataclass
class AnyList(Pattern):
    """Returns whether the key evaluates to true anywhere in the list"""

    lst: AST
    key: AST

    def compose(self):
        return Apply(
            PLambda(
                ["op"],
                RecFun(
                    PLambda(
                        ["f", "xs"],
                        IteNullList(
                            PVar("xs"),
                            Bool(False),
                            Ite(
                                Apply(PVar("op"), HeadList(PVar("xs"))),
                                Bool(True),
                                Apply(
                                    PVar("f"),
                                    PVar("f"),
                                    TailList(PVar("xs")),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
            self.key,
            self.lst,
        )


@dataclass
class AllList(Pattern):
    """Returns whether the key evaluates to true everywhere in the list"""

    lst: AST
    key: AST

    def compose(self):
        return Apply(
            PLambda(
                ["op"],
                RecFun(
                    PLambda(
                        ["f", "xs"],
                        IteNullList(
                            PVar("xs"),
                            Bool(True),
                            Ite(
                                Apply(PVar("op"), HeadList(PVar("xs"))),
                                Apply(
                                    PVar("f"),
                                    PVar("f"),
                                    TailList(PVar("xs")),
                                ),
                                Bool(False),
                            ),
                        ),
                    ),
                ),
            ),
            self.key,
            self.lst,
        )


@dataclass
class FilterList(Pattern):
    """Apply a filter function on each element in a list (throws out all that evaluate to false)"""

    lst: AST
    k: AST
    empty_list: AST = field(default_factory=EmptyDataList)

    def compose(self):
        return Apply(
            PLambda(
                ["op"],
                RecFun(
                    PLambda(
                        ["filter", "xs"],
                        IteNullList(
                            PVar("xs"),
                            self.empty_list,
                            PLet(
                                [("head", HeadList(PVar("xs")))],
                                Ite(
                                    Apply(PVar("op"), PVar("head")),
                                    PrependList(
                                        PVar("head"),
                                        Apply(
                                            PVar("filter"),
                                            PVar("filter"),
                                            TailList(PVar("xs")),
                                        ),
                                    ),
                                    Apply(
                                        PVar("filter"),
                                        PVar("filter"),
                                        TailList(PVar("xs")),
                                    ),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
            self.k,
            self.lst,
        )


@dataclass
class MapFilterList(Pattern):
    """
    Apply a filter and a map function on each element in a list (throws out all that evaluate to false)
    Performs only a single pass and is hence much more efficient than filter + map
    """

    lst: AST
    filter_op: AST
    map_op: AST
    empty_list: AST = field(default_factory=EmptyDataList)

    def compose(self):
        return Apply(
            PLambda(
                ["filter", "map"],
                RecFun(
                    PLambda(
                        ["filtermap", "xs"],
                        IteNullList(
                            PVar("xs"),
                            self.empty_list,
                            PLet(
                                [("head", HeadList(PVar("xs")))],
                                Ite(
                                    Apply(PVar("filter"), PVar("head")),
                                    PrependList(
                                        Apply(PVar("map"), PVar("head")),
                                        Apply(
                                            PVar("filtermap"),
                                            PVar("filtermap"),
                                            TailList(PVar("xs")),
                                        ),
                                    ),
                                    Apply(
                                        PVar("filtermap"),
                                        PVar("filtermap"),
                                        TailList(PVar("xs")),
                                    ),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
            self.filter_op,
            self.map_op,
            self.lst,
        )


@dataclass
class LengthList(Pattern):
    lst: AST

    def compose(self):
        return FoldList(
            self.lst, PLambda(["a", "_"], AddInteger(PVar("a"), Integer(1))), Integer(0)
        )


@dataclass
class TakeList(Pattern):
    """Take the first n elements of list l"""

    lst: AST
    n: AST
    empty_list: AST = field(default_factory=EmptyDataList)

    def compose(self):
        return Apply(
            RecFun(
                PLambda(
                    ["take", "xs", "n"],
                    IteNullList(
                        PVar("xs"),
                        self.empty_list,
                        Ite(
                            LessThanEqualsInteger(PVar("n"), Integer(0)),
                            self.empty_list,
                            PrependList(
                                HeadList(PVar("xs")),
                                Apply(
                                    PVar("take"),
                                    PVar("take"),
                                    TailList(PVar("xs")),
                                    SubtractInteger(PVar("n"), Integer(1)),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
            self.lst,
            self.n,
        )


@dataclass
class DropList(Pattern):
    """Drop the first n elements of list l"""

    lst: AST
    n: AST
    empty_list: AST = field(default_factory=EmptyDataList)

    def compose(self):
        return Apply(
            RecFun(
                PLambda(
                    ["drop", "xs", "n"],
                    IteNullList(
                        PVar("xs"),
                        self.empty_list,
                        Ite(
                            LessThanEqualsInteger(PVar("n"), Integer(0)),
                            PVar("xs"),
                            Apply(
                                PVar("drop"),
                                PVar("drop"),
                                TailList(PVar("xs")),
                                SubtractInteger(PVar("n"), Integer(1)),
                            ),
                        ),
                    ),
                ),
            ),
            self.lst,
            self.n,
        )


@dataclass
class SliceList(Pattern):
    """Drop the first i elements of list l and take the remaining j elements"""

    i: AST
    j: AST
    lst: AST
    empty_list: AST = field(default_factory=EmptyDataList)

    def compose(self):
        return TakeList(
            DropList(self.lst, self.i, self.empty_list), self.j, self.empty_list
        )


# Data Utils


@dataclass
class Constructor(Pattern):
    d: AST

    def compose(self):
        return FstPair(UnConstrData(self.d))


@dataclass
class Fields(Pattern):
    d: AST

    def compose(self):
        return SndPair(UnConstrData(self.d))


@dataclass
class NthField(Pattern):
    d: AST
    n: AST

    def compose(self):
        return IndexAccessList(Fields(self.d), self.n)


def ConstantNthField(d: AST, i: int):
    return ConstantIndexAccessList(Fields(d), i)


def ConstantNthFieldFast(d: AST, i: int):
    return ConstantIndexAccessListFast(Fields(d), i)


@dataclass
class NoneData(Pattern):
    def compose(self):
        return ConstrData(Integer(0), EmptyDataList())


@dataclass
class SomeData(Pattern):
    x: AST
    # Note: x must be of type data!

    def compose(self):
        return ConstrData(Integer(1), SingleDataList(self.x))


# Choose Utils


def DelayedChooseData(
    d: AST,
    constr_branch: AST,
    map_branch: AST,
    list_branch: AST,
    int_branch: AST,
    bytestring_branch: AST,
):
    # Careful: cannot patternize because of delay/force
    return Force(
        ChooseData(
            d,
            Delay(constr_branch),
            Delay(map_branch),
            Delay(list_branch),
            Delay(int_branch),
            Delay(bytestring_branch),
        )
    )


# concatenation utils


def _concat(append, empty):
    def f(*ss: typing.List[AST]):
        if not ss:
            return empty
        c = ss[-1]
        for s in reversed(ss[:-1]):
            c = append(s, c)
        return c

    return f


@dataclass
class AppendList(Pattern):
    xs: AST
    ys: AST

    def compose(self):
        return Apply(
            RecFun(
                PLambda(
                    ["append", "xs", "ys"],
                    IteNullList(
                        PVar("xs"),
                        PVar("ys"),
                        PrependList(
                            HeadList(PVar("xs")),
                            Apply(
                                PVar("append"),
                                PVar("append"),
                                TailList(PVar("xs")),
                                PVar("ys"),
                            ),
                        ),
                    ),
                )
            ),
            self.xs,
            self.ys,
        )


ConcatString = _concat(AppendString, Text(""))
ConcatByteString = _concat(AppendByteString, ByteString(b""))


def ConcatList(sample_value: uplc_ast.Constant):
    return _concat(AppendList, EmptyList(sample_value))



import dataclasses
import uuid
from collections import defaultdict
from copy import deepcopy
from functools import lru_cache
from typing import Type
from graphlib import TopologicalSorter
from ordered_set import OrderedSet

from .. import PVar, PLambda, PLet, Ite
from ..pluthon_ast import Pattern, Program, Apply, Force, Delay, Var, Lambda
from ..util import NodeTransformer, NodeVisitor, iter_fields


class EvaluatedVariableCollector(NodeVisitor):
    def __init__(self):
        self.evaluated_variables = OrderedSet()

    def visit_Var(self, node: Var):
        self.evaluated_variables.add(node.name)


class ConditionallyEvaluatedVariableCollector(NodeVisitor):
    def __init__(self):
        self.conditionally_evaluated_variables = OrderedSet()

    def visit_Ite(self, node: Ite):
        then_branch_evaluated_variables_collector = EvaluatedVariableCollector()
        then_branch_evaluated_variables_collector.visit(node.t)
        self.conditionally_evaluated_variables.update(
            then_branch_evaluated_variables_collector.evaluated_variables
        )
        else_branch_evaluated_variables_collector = EvaluatedVariableCollector()
        else_branch_evaluated_variables_collector.visit(node.e)
        self.conditionally_evaluated_variables.update(
            else_branch_evaluated_variables_collector.evaluated_variables
        )

    def visit_Delay(self, node: Delay):
        evaluated_variables_collector = EvaluatedVariableCollector()
        evaluated_variables_collector.visit(node.x)
        self.conditionally_evaluated_variables.update(
            evaluated_variables_collector.evaluated_variables
        )

    def visit_Lambda(self, node: Lambda):
        evaluated_variables_collector = EvaluatedVariableCollector()
        evaluated_variables_collector.visit(node.term)
        self.conditionally_evaluated_variables.update(
            evaluated_variables_collector.evaluated_variables
        )


@lru_cache()
def conditionally_evaluated_params(pattern_class: Type[Pattern]):
    """
    taint analysis to figure out if parameters to a pattern are involved conditionally -> if so, we need to wrap them in a delay
    """
    # we generate unique identifiers for all parameters of the pattern
    # to detect which of them are used conditionally
    fields = dataclasses.fields(pattern_class)
    uuid_map = {f.name: PVar(f"{f.name}_{uuid.uuid4().hex}").name for f in fields}
    if fields:
        term = pattern_class(*[Var(uuid_map[f.name]) for f in fields]).compose()
    else:
        term = pattern_class().compose()
    conditionally_evaluated_variables_collector = (
        ConditionallyEvaluatedVariableCollector()
    )
    conditionally_evaluated_variables_collector.visit(term)
    return {
        f.name
        for f in fields
        if uuid_map[f.name]
        in conditionally_evaluated_variables_collector.conditionally_evaluated_variables
    }


class PatternCollector(NodeVisitor):
    def __init__(self):
        self.patterns = dict()

    def visit(self, node):
        """Visit a node."""
        if isinstance(node, Pattern):
            # Patterns are special
            # we collect them here and later add them in reverse order
            # after subpatterns are added recursively
            # this ensures that the outermost pattern is added last
            node_type = type(node)
            self.patterns[node_type] = None
            res = self.visit(node.compose())
        else:
            method = "visit_" + node.__class__.__name__
            visitor = getattr(self, method, self.generic_visit)
            res = visitor(node)
        return res


class PatternDepBuilder(NodeVisitor):
    def __init__(self):
        self.pattern_deps = dict()
        self.pattern_occurrences = defaultdict(int)

    def patterns_in_dep_order(self):
        """
        Returns the patterns in dependency order, i.e. the ones that are used in other patterns are defined first
        """
        ts = TopologicalSorter(self.pattern_deps)
        return ts.static_order()

    def visit(self, node):
        """Visit a node."""
        if isinstance(node, Pattern):
            # Patterns are special
            node_type = type(node)
            self.pattern_occurrences[node_type] += 1
            # compose but without actual variables to avoid collecting patterns that are passed into the pattern
            subpattern_collector = PatternCollector()
            subpattern_collector.visit(make_abstract_function(node_type))
            subpatterns = subpattern_collector.patterns
            self.pattern_deps.setdefault(node_type, OrderedSet()).update(subpatterns)
            res = self.visit(node.compose())
        else:
            method = "visit_" + node.__class__.__name__
            visitor = getattr(self, method, self.generic_visit)
            res = visitor(node)
        return res


@lru_cache()
def make_abstract_function(pattern_class: Type[Pattern]):
    fields = dataclasses.fields(pattern_class)
    cep = conditionally_evaluated_params(pattern_class)
    if fields:
        return PLambda(
            [f.name for f in fields],
            pattern_class(
                *[
                    Force(PVar(f.name)) if f.name in cep else PVar(f.name)
                    for f in fields
                ]
            ).compose(),
        )
    else:
        return pattern_class().compose()


def make_abstract_function_name(pattern_class: Type[Pattern]):
    return f"p_{pattern_class.__name__}"


class OncePatternReplacer(NodeTransformer):
    """
    Replaces the innermost pattern in terms of dependencies, i.e. the one that has to be defined
    last in topological order
    """

    unfold_pattern_class = None

    def visit(self, node):
        """Visit a node."""
        if self.unfold_pattern_class is not None and isinstance(
            node, self.unfold_pattern_class
        ):
            # Patterns are special
            if self.unfold_pattern_occurrences > 1:
                pattern_var = PVar(make_abstract_function_name(type(node)))
                fields = list(iter_fields(node))
                cep = conditionally_evaluated_params(type(node))
                if fields:
                    node = Apply(
                        pattern_var,
                        *(
                            Delay(field) if name in cep else field
                            for name, field in iter_fields(node)
                        ),
                    )
                else:
                    node = pattern_var
            else:
                node = node.compose()
        method = "visit_" + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def visit_Program(self, node: Program):
        pattern_collector = PatternDepBuilder()
        pattern_collector.visit(node)
        pattern_classes = list(pattern_collector.patterns_in_dep_order())
        if pattern_classes:
            self.unfold_pattern_class = pattern_classes[-1]
            self.unfold_pattern_occurrences = pattern_collector.pattern_occurrences[
                self.unfold_pattern_class
            ]
            if self.unfold_pattern_occurrences > 1:
                # if the pattern occurs more than once, we need to define it as a function
                # otherwise we can just inline it
                term = PLet(
                    [
                        (
                            make_abstract_function_name(self.unfold_pattern_class),
                            self.visit(
                                deepcopy(
                                    make_abstract_function(self.unfold_pattern_class)
                                )
                            ),
                        ),
                    ],
                    self.visit(node.prog),
                )
            else:
                term = self.visit(node.prog)
        else:
            term = self.visit(node.prog)
        self.unfold_pattern_class = None
        return Program(
            version=node.version,
            prog=term,
        )


class AllPatternReplacer(NodeTransformer):
    def visit(self, node):
        """Visit a node."""
        if isinstance(node, Pattern):
            # Patterns are special
            pattern_var = PVar(make_abstract_function_name(type(node)))
            fields = list(iter_fields(node))
            cep = conditionally_evaluated_params(type(node))
            if fields:
                node = Apply(
                    pattern_var,
                    *(
                        Delay(field) if name in cep else field
                        for name, field in iter_fields(node)
                    ),
                )
            else:
                node = pattern_var
        method = "visit_" + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def visit_Program(self, node: Program):
        pattern_collector = PatternDepBuilder()
        pattern_collector.visit(node)
        pattern_classes = list(pattern_collector.patterns_in_dep_order())
        if pattern_classes:
            term = PLet(
                [
                    (
                        make_abstract_function_name(pattern_class),
                        self.visit(deepcopy(make_abstract_function(pattern_class))),
                    )
                    for pattern_class in pattern_classes
                ],
                self.visit(node.prog),
            )
        else:
            term = node.prog
        return Program(
            version=node.version,
            prog=term,
        )



from pluthon.pluthon_ast import Apply, BuiltIn, Force, Text
from pluthon.util import NodeTransformer
from uplc import ast as uplc_ast


class RemoveTrace(NodeTransformer):
    """
    Replaces Trace with just the argument
    """

    def visit_Apply(self, node: Apply):
        if (
            isinstance(node.f, Force)
            and isinstance(node.f.x, BuiltIn)
            and node.f.x.builtin == uplc_ast.BuiltInFun.Trace
            # only remove traces that trace a constant string
            # otherwise there might be side effects we are removing
            and isinstance(node.xs[0], Text)
        ):
            return self.generic_visit(node.xs[1])
        return self.generic_visit(node)
